/* -*- c-file-style: "java"; indent-tabs-mode: nil -*-
 * 
 * distcc -- A simple distributed compiler system
 * $Header: /data/cvs/distcc/src/lock.c,v 1.4 2002/10/07 01:35:19 mbp Exp $ 
 *
 * Copyright (C) 2002 by Martin Pool <mbp@samba.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */


#define _GNU_SOURCE

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>

#include <sys/stat.h>
#include <sys/file.h>

#include "distcc.h"
#include "trace.h"
#include "util.h"
#include "hosts.h"
#include "tempfile.h"
#include "lock.h"
#include "io.h"


const struct dcc_hostdef _dcc_local = {
    DCC_MODE_LOCAL,
    NULL,
    "localhost",
    0,
    NULL,
    NULL    
};

const struct dcc_hostdef *dcc_hostdef_local = &_dcc_local;


/* TODO: Lock files ought to incorporate port. */

/* TODO: Lock files ought to also incorporate IP address, so that we
 * can use multi-A records for round-robin assignment. */

static char * dcc_make_lock_filename(const char *host, int iter)
{
    int need_len;
    char * buf;
    const char *tempdir;

    tempdir = dcc_get_tempdir();
    need_len = strlen(tempdir) + 6 + strlen(host) + 1 + 7 + 1;
    buf = malloc(need_len);
    if (snprintf(buf, need_len, "%s/lock_%s_%07d", tempdir, host, iter)
        != need_len - 1) {
        rs_fatal("wrong length??");
    }
    return buf;
}


/**
 * Get an exclusive, non-blocking lock on a file using whatever method
 * is available on this system.
 *
 * @return 0 if we got the lock; -1 with EAGAIN if the file is already
 * locked.
 **/
static int sys_lock(int fd)
{
#if defined(F_SETLK)
    struct flock lockparam;

    lockparam.l_type = F_WRLCK;
    lockparam.l_whence = SEEK_SET;
    lockparam.l_start = 0;
    lockparam.l_len = 0;        /* whole file */
    
    return fcntl(fd, F_SETLK, &lockparam);
#elif defined(HAVE_FLOCK)
    return flock(fd, LOCK_EX|LOCK_NB);
#elif defined(HAVE_LOCKF)
    return lockf(fd, F_TLOCK, 0);
#else
#  error "No supported lock method.  Please port this code."
#endif
}


int dcc_try_lock_host(const struct dcc_hostdef *host, int iter)
{
    const char *tempdir;
    char *fname;
    int fd = -1;
    int ret;

    tempdir = dcc_get_tempdir();
    fname = dcc_make_lock_filename(host->hostname, iter);

    /* Create if it doesn't exist.  We don't actually do anything with
     * the file except lock it.*/
    if ((fd = open(fname, O_WRONLY|O_CREAT, 0600)) == -1
        && errno != EEXIST) {
        rs_log_error("failed to creat %s: %s", fname, strerror(errno));
        goto bomb;
    }

    ret = sys_lock(fd);
    
    if (ret != -1) {
        rs_trace("locked %s", fname);
        free(fname);
        return 0;
    } else
        switch (errno) {
#ifdef EWOULDBLOCK
        case EWOULDBLOCK:
#endif
#if defined(EAGAIN) && EAGAIN != EWOULDBLOCK
        case EAGAIN:
#endif
            rs_trace("%s already locked", fname);
            break;
        default:
            rs_log_error("flock %s failed: %s", fname, strerror(errno));
            break;
    }

    bomb:
    if (fd != -1)
        dcc_close(fd);
    free(fname);
    return -1;
}


/**
 * Lock localhost.  Used to get the right balance of jobs when some of
 * them must be local.
 **/
int dcc_lock_local(void)
{
    int i_try;
    
    /* FIXME: Should use a function that blocks until a lock can be
     * taken. */

    for (i_try = 0; i_try < 50; i_try++) {
        if (dcc_try_lock_host(dcc_hostdef_local, i_try) == 0) {
            rs_trace("got lock on localhost");
            return 0;
        }
    }
    
    return -1;                  /* but go anyhow */
}

                   
